from typing import Optional, Dict


Fore = str
Back = str
Style = str
CodeMap = Dict[str, Dict[str, int]]

_no_color = False


code_map: CodeMap = {
    "fore": {
        "black": 30,
        "red": 31,
        "green": 32,
        "yellow": 33,
        "blue": 34,
        "magenta": 35,
        "cyan": 36,
        "white": 37,
    },
    "back": {
        "black": 40,
        "red": 41,
        "green": 42,
        "yellow": 43,
        "blue": 44,
        "magenta": 45,
        "cyan": 46,
        "white": 47,
    },
    "style": {
        "reset": 0,
        "bold": 1,
        "italic": 3,
        "underline": 4,
        "defaultfg": 39,
        "defaultbg": 49,
    },
}


def colored_string(
    string: str, fore: Optional[Fore] = None, back: Optional[Back] = None, style: Optional[Style] = None
) -> str:
    if _no_color:
        return string
    template = "\033[{code}m"
    result = ""
    if fore is not None:
        result += template.format(code=code_map["fore"][fore])
    if back is not None:
        result += template.format(code=code_map["back"][back])
    if style is not None:
        result += template.format(code=code_map["style"][style])
    result += string
    result += template.format(code=code_map["style"]["reset"])
    return result


def set_no_color():
    global _no_color
    _no_color = True



from ..tools import spider
from .utils import MaxRetry


@MaxRetry(2)
def is_vip() -> bool:
    info_api = "https://api.bilibili.com/x/web-interface/nav"
    res_json = spider.get(info_api, timeout=3).json()
    res_json_data = res_json.get("data")
    if res_json_data.get("vipStatus") == 1:
        return True
    return False



import os
import random
from typing import Callable, List, Tuple, Union

import requests

from ..handlers.base import Handler
from ..utils.console.logger import Logger
from ..utils.crawler import Crawler


class RemoteFile(Handler):
    """远程文件类

    网络 url 与本地文件的绑定，可调用 download 进行下载
    download 支持断点续传
    """

    before_download: Callable[..., None]
    downloaded: Callable[..., None]
    before_update: Callable[..., None]
    updated: Callable[..., None]

    def __init__(
        self, url: str, local_path: str, mirrors: List[str] = [], range: Tuple[int, Union[int, str]] = (0, "")
    ):
        super().__init__(["before_download", "before_update", "updated", "downloaded"])
        self.url = url
        self.mirrors = mirrors
        self.path = local_path
        self.name = os.path.split(self.path)[-1]
        self.tmp_path = self.path + ".dl"
        self.size = self.get_local_size()
        self.range = range

    def get_local_size(self) -> int:
        """ 通过 os.path.getsize 获取本地文件大小 """
        try:
            if os.path.exists(self.tmp_path):
                size = os.path.getsize(self.tmp_path)
            elif os.path.exists(self.path):
                size = os.path.getsize(self.path)
            else:
                size = 0
        except FileNotFoundError:
            size = 0
        return size

    def download(self, thread_spider: Crawler, stream: bool = True, chunk_size: int = 1024):
        """[summary]

        Args:
            thread_spider (requests.Session): 线程全局下载器，由线程池管理并传入，每个线程拥有一个
            stream (bool, optional): 是否启用流式下载. Defaults to True.
            chunk_size (int, optional): 块大小. Defaults to 1024.
        """
        spider = thread_spider
        self.before_download(self)
        if not os.path.exists(self.path):
            downloaded = False
            while not downloaded:
                # 设置 headers
                headers = dict(spider.headers)
                headers["Range"] = "bytes={}-{}".format(self.size + self.range[0], self.range[1])
                url = random.choice([self.url] + self.mirrors) if self.mirrors else self.url

                try:
                    # 尝试建立连接
                    res = spider.get(url, stream=stream, headers=headers, timeout=(5, 10))
                    # 下载到临时路径
                    with open(self.tmp_path, "ab") as f:
                        if stream:
                            for chunk in res.iter_content(chunk_size=chunk_size):
                                if not chunk:
                                    break
                                self.before_update(self)
                                f.write(chunk)
                                self.size += len(chunk)
                                self.updated(self)
                        else:
                            f.write(res.content)
                    # size 检验，因为有时明明没下完仍然会停止下载
                    if self.range[1] and (self.range[1] - self.range[0] + 1 != self.size):
                        downloaded = False
                    else:
                        downloaded = True
                except requests.exceptions.RequestException:
                    Logger.warning("文件 {} 下载超时，正在重试...".format(self.name))

            # 从临时文件迁移，并删除临时文件
            if os.path.exists(self.path):
                os.remove(self.path)
            os.rename(self.tmp_path, self.path)

        self.downloaded(self)



from typing import List, Union

from ..utils.console.logger import Logger


class Status:
    """ 多层次状态管理类 """

    def __init__(self, parent: Union["Status", None] = None, children: List["Status"] = []):
        self.parent = None
        self.children = []
        if parent is not None:
            self.set_parent(parent)
        if children:
            self.add_children(children)

    def add_child(self, child: "Status"):
        self.children.append(child)
        child.parent = self

    def set_parent(self, parent: "Status"):
        parent.add_child(self)

    def add_children(self, children: List["Status"]):
        for child in children:
            self.add_child(child)

    @property
    def is_leaf(self) -> bool:
        return not self.children

    @property
    def is_root(self) -> bool:
        return self.parent is None


class DownloaderStatus(Status):
    """下载状态类"""

    def __init__(self, parent: Union["Status", None] = None, children: List["Status"] = []):
        super().__init__(parent=parent, children=children)
        self.__total_size: int = 0
        self.__size: int = 0
        self.__downloading: bool = False
        self.__downloaded: bool = False
        self.__merging: bool = False
        self.__merged: bool = False

    @property
    def total_size(self) -> int:
        if self.is_leaf:
            return self.__total_size
        else:
            return sum([child.total_size for child in self.children])

    @total_size.setter
    def total_size(self, value: int):
        if self.is_leaf:
            self.__total_size = value
        else:
            Logger.error("无法设定非叶子结点的 total_size")

    @property
    def size(self) -> int:
        if self.is_leaf:
            if self.downloaded:
                return self.total_size
            return self.__size
        else:
            return sum([child.size for child in self.children])

    @size.setter
    def size(self, value: int):
        if self.is_leaf:
            self.__size = value
        else:
            Logger.error("无法设定非叶子结点的 size")

    @property
    def downloading(self) -> bool:
        if self.is_leaf:
            return self.__downloading
        else:
            return any([child.downloading for child in self.children])

    @downloading.setter
    def downloading(self, value: bool):
        if self.is_leaf:
            self.__downloading = value
        else:
            if value:
                Logger.error("无法设定非叶子结点的 downloading 为 True")
            else:
                for child in self.children:
                    child.downloading = False

    @property
    def downloaded(self) -> bool:
        if self.is_leaf:
            return self.__downloaded
        else:
            return all([child.downloaded for child in self.children])

    @downloaded.setter
    def downloaded(self, value: bool):
        if self.is_leaf:
            self.__downloaded = value
        else:
            if value:
                for child in self.children:
                    child.downloaded = True
            else:
                Logger.error("无法设定非叶子结点的 downloaded 为 False")

    @property
    def merging(self) -> bool:
        if self.is_leaf:
            return self.__merging
        else:
            return any([child.merging for child in self.children])

    @merging.setter
    def merging(self, value: bool):
        if self.is_leaf:
            self.__merging = value
        else:
            if value:
                # 由于合并是共同进行的，所以可以由父结点来设置
                for child in self.children:
                    child.merging = True
            else:
                for child in self.children:
                    child.merging = False

    @property
    def merged(self) -> bool:
        if self.is_leaf:
            return self.__merged
        else:
            return all([child.merged for child in self.children])

    @merged.setter
    def merged(self, value: bool):
        if self.is_leaf:
            self.__merged = value
        else:
            if value:
                for child in self.children:
                    child.merged = True
            else:
                Logger.error("无法设定非叶子结点的 merged 为 False")



import os
from typing import Dict

import requests


class DownloadFailureError(Exception):
    pass


class Crawler(requests.Session):

    header = {
        "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 11_3_0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/89.0.4389.128 Safari/537.36",
    }

    def __init__(self):
        super().__init__()
        requests.packages.urllib3.disable_warnings()
        self.headers.update(Crawler.header)
        self.verify = False

    def set_cookies(self, cookies: Dict[str, str]):
        """传入一个字典，用于设置 cookies"""

        self.cookies_dict = cookies
        requests.utils.add_dict_to_cookiejar(self.cookies, cookies)

    def download_bin(
        self, url: str, file_path: str, stream: bool = True, chunk_size: int = 1024, **kw: Dict[str, str]
    ) -> None:
        """下载二进制文件"""

        res = self.get(url, stream=stream, **kw)
        tmp_path = file_path + ".t"
        try:
            with open(tmp_path, "wb") as f:
                if stream:
                    for chunk in res.iter_content(chunk_size=chunk_size):
                        if not chunk:
                            break
                        f.write(chunk)
                else:
                    f.write(res.content)
        except:
            os.remove(tmp_path)
            raise DownloadFailureError()
        if os.path.exists(file_path):
            with open(tmp_path, "rb") as fr:
                with open(file_path, "wb") as fw:
                    fw.write(fr.read())
            os.remove(tmp_path)
        else:
            os.rename(tmp_path, file_path)

    def download_text(self, url: str, file_path: str, **kw: Dict[str, str]) -> None:
        """下载文本，以 UTF-8 编码保存文件"""

        res = self.get(url, **kw)
        res.encoding = res.apparent_encoding
        with open(file_path, "w", encoding="utf_8") as f:
            f.write(res.text)

    def clone(self):
        new_one = self.__class__()
        new_one.set_cookies(self.cookies_dict)
        new_one.trust_env = self.trust_env
        return new_one


class BililiCrawler(Crawler):

    header = {
        "User-Agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.135 Safari/537.36",
        "Referer": "https://www.bilibili.com",
    }

    def __init__(self):
        super().__init__()
        self.headers.update(BililiCrawler.header)



import os
import random
import subprocess
import shutil

from typing import List


class FFmpegNotFoundError(Exception):
    def __init__(self):
        super().__init__("请配置正确的 FFmpeg 路径")


class FFmpeg:
    """
    @refs : https://github.com/soimort/you-get
    """

    def __init__(self, ffmpeg_path: str = "ffmpeg"):
        try:
            if subprocess.run([ffmpeg_path], stdout=subprocess.PIPE, stderr=subprocess.PIPE).returncode != 1:
                raise FFmpegNotFoundError()
        except FileNotFoundError:
            raise FFmpegNotFoundError()
        self.path = os.path.normpath(ffmpeg_path)
        tmp_dir = os.path.join(os.path.dirname(ffmpeg_path), "tmp")
        if not os.path.exists(tmp_dir):
            os.makedirs(tmp_dir)
        self.tmp_dir = os.path.normpath(tmp_dir)

    def __del__(self):
        if hasattr(self, "tmp_dir") and os.path.exists(self.tmp_dir):
            shutil.rmtree(self.tmp_dir)

    def exec(self, params: List[str]):
        """ 调用 ffmpeg """
        cmd = [self.path]
        cmd.extend(params)
        return subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    def convert(self, input_path: str, output_path: str) -> None:
        """ 视频格式转换 """

        # fmt: off
        params = [
            "-i", input_path,
            "-c", "copy",
            "-map", "0",
            "-y",
            output_path
        ]
        # fmt: on
        self.exec(params)

    def join_videos(self, video_path_list: List[str], output_path: str) -> None:
        """ 将视频拼接起来 """

        concat_list_path = os.path.join(self.tmp_dir, "concat_list_{:04}.tmp".format(random.randint(0, 9999))).replace(
            "\\", "/"
        )
        with open(concat_list_path, "w", encoding="utf-8") as f:
            for video_path in video_path_list:
                if os.path.isfile(video_path):
                    video_relpath = os.path.relpath(video_path, start=self.tmp_dir)
                    f.write("file '{}'\n".format(video_relpath))
        # fmt: off
        params = [
            "-f", "concat",
            "-safe", "-1",
            "-i", concat_list_path,
            "-c", "copy",
            "-y",
            output_path
        ]
        # fmt: on
        self.exec(params)
        os.remove(concat_list_path)

    def join_video_audio(self, video_path: str, audio_path: str, output_path: str) -> None:
        """ 将视频和音频合并 """

        # fmt: off
        params = [
            "-i", video_path,
            "-i", audio_path,
            "-codec", "copy",
            "-y",
            output_path
        ]
        # fmt: on

        self.exec(params)



import re

from ..api.exceptions import ArgumentsError, CannotDownloadError, IsPreviewError, UnknownTypeError, UnsupportTypeError
from ..quality import Media, gen_quality_sequence, video_quality_map
from ..tools import spider
from ..utils.base import touch_url
from .utils import MaxRetry


@MaxRetry(2)
def get_season_id(media_id: str) -> str:
    home_url = "https://api.bilibili.com/pgc/review/user?media_id={media_id}".format(media_id=media_id)
    res_json = spider.get(home_url).json()
    return str(res_json["result"]["media"]["season_id"])


@MaxRetry(2)
def get_bangumi_title(media_id: str = "", season_id: str = "", episode_id: str = "") -> str:
    if not (media_id or season_id or episode_id):
        raise ArgumentsError("media_id", "season_id", "episode_id")
    title = "呐，我也不知道是什么标题呢～"
    if media_id:
        home_url = "https://www.bilibili.com/bangumi/media/md{media_id}".format(media_id=media_id)
        res = spider.get(home_url, timeout=3)
        regex_title = re.compile(r'<span class="media-info-title-t">(.*?)</span>')
        if match_obj := regex_title.search(res.text):
            title = match_obj.group(1)
    elif season_id or episode_id:
        if season_id:
            play_url = "https://www.bilibili.com/bangumi/play/ss{season_id}".format(season_id=season_id)
        else:
            play_url = "https://www.bilibili.com/bangumi/play/ep{episode_id}".format(episode_id=episode_id)
        res = spider.get(play_url, timeout=3)
        regex_title = re.compile(r'<a href=".+" target="_blank" title="(.*?)" class="media-title">(?P<title>.*?)</a>')
        if match_obj := regex_title.search(res.text):
            title = match_obj.group("title")
    return title


@MaxRetry(2)
def get_bangumi_list(episode_id: str = "", season_id: str = "", with_section: bool = False):
    if not (season_id or episode_id):
        raise ArgumentsError("season_id", "episode_id")
    list_api = "http://api.bilibili.com/pgc/view/web/season?season_id={season_id}&ep_id={episode_id}"
    res = spider.get(list_api.format(episode_id=episode_id, season_id=season_id), timeout=3)
    result = res.json()["result"]
    section_episodes = []

    if with_section and result.get("section", []):
        for section in result["section"]:
            section_episodes += section["episodes"]

    return [
        {
            "id": i + 1,
            "name": " ".join(
                [
                    "第{}话".format(item["title"]) if re.match(r"^\d*\.?\d*$", item["title"]) else item["title"],
                    item["long_title"],
                ]
            ),
            "cid": str(item["cid"]),
            "episode_id": str(item["id"]),
            "avid": str(item["aid"]),
            "bvid": item["bvid"],
        }
        for i, item in enumerate(result["episodes"] + section_episodes)
    ]


@MaxRetry(2)
def get_bangumi_playurl(
    avid: str = "",
    bvid: str = "",
    episode_id: str = "",
    cid: str = "",
    quality: int = 127,
    audio_quality: int = 30280,
    type: str = "dash",
):
    video_quality_sequence = gen_quality_sequence(quality, type=Media.VIDEO)
    audio_quality_sequence = gen_quality_sequence(audio_quality, type=Media.AUDIO)
    play_api = "https://api.bilibili.com/pgc/player/web/playurl?avid={avid}&bvid={bvid}&ep_id={episode_id}&cid={cid}&qn={quality}"
    if type == "flv":
        touch_message = spider.get(
            play_api.format(avid=avid, bvid=bvid, episode_id=episode_id, cid=cid, quality=80), timeout=3
        ).json()
        if touch_message["code"] != 0:
            raise CannotDownloadError(touch_message["code"], touch_message["message"])
        if touch_message["result"]["is_preview"] == 1:
            raise IsPreviewError()

        accept_quality = touch_message["result"]["accept_quality"]
        for quality in video_quality_sequence:
            if quality in accept_quality:
                break

        play_url = play_api.format(avid=avid, bvid=bvid, episode_id=episode_id, cid=cid, quality=quality)
        res = spider.get(play_url, timeout=3)

        return [
            {
                "id": i + 1,
                "url": segment["url"],
                "mirrors": segment["backup_url"],
                "quality": quality,
                "height": video_quality_map[quality]["height"],
                "width": video_quality_map[quality]["width"],
                "size": segment["size"],
                "type": "flv_segment",
            }
            for i, segment in enumerate(res.json()["result"]["durl"])
        ]
    elif type == "dash":
        result = []
        play_api_dash = play_api + "&fnver=0&fnval=2000&fourk=1"
        play_info = spider.get(
            play_api_dash.format(
                avid=avid,
                bvid=bvid,
                episode_id=episode_id,
                cid=cid,
                quality=video_quality_sequence[0],
            ),
            timeout=3,
        ).json()

        if play_info["code"] != 0:
            raise CannotDownloadError(play_info["code"], play_info["message"])
        if play_info["result"].get("dash") is None:
            raise UnsupportTypeError("dash")
        if play_info["result"]["is_preview"] == 1:
            raise IsPreviewError()

        accept_video_quality = set([video["id"] for video in play_info["result"]["dash"]["video"]])
        for video_quality in video_quality_sequence:
            if video_quality in accept_video_quality:
                break
        else:
            video_quality = 127

        accept_audio_quality = set([audio["id"] for audio in play_info["result"]["dash"]["audio"]])
        for audio_quality in audio_quality_sequence:
            if audio_quality in accept_audio_quality:
                break
        else:
            audio_quality = 30280

        if play_info["result"]["dash"]["video"]:
            videos = play_info["result"]["dash"]["video"]
            for video in videos:
                if video["id"] == video_quality:
                    result.append(
                        {
                            "id": 1,
                            "url": video["base_url"],
                            "mirrors": video["backup_url"],
                            "quality": video_quality,
                            "height": video["height"],
                            "width": video["width"],
                            "size": touch_url(video["base_url"], spider)[0],
                            "type": "dash_video",
                        }
                    )
                    break
        if play_info["result"]["dash"]["audio"]:
            audios = play_info["result"]["dash"]["audio"]
            for audio in audios:
                if audio["id"] == audio_quality:
                    result.append(
                        {
                            "id": 2,
                            "url": audio["base_url"],
                            "mirrors": audio["backup_url"],
                            "quality": audio_quality,
                            "height": None,
                            "width": None,
                            "size": touch_url(audio["base_url"], spider)[0],
                            "type": "dash_audio",
                        }
                    )
                    break
        return result
    elif type == "mp4":
        raise UnsupportTypeError("mp4")
    else:
        raise UnknownTypeError(type)


@MaxRetry(2)
def get_bangumi_subtitle(avid: str = "", bvid: str = "", cid: str = ""):
    if not (avid or bvid):
        raise ArgumentsError("avid", "bvid")
    subtitle_api = "https://api.bilibili.com/x/player/v2?cid={cid}&aid={avid}&bvid={bvid}"
    subtitle_url = subtitle_api.format(avid=avid, bvid=bvid, cid=cid)
    subtitles_info = spider.get(subtitle_url, timeout=3).json()["data"]["subtitle"]
    return [
        # fmt: off
        {
            "lang": sub_info["lan_doc"],
            "lines": spider.get("https:" + sub_info["subtitle_url"], timeout=(3, 10)).json()["body"]
        }
        for sub_info in subtitles_info["subtitles"]
        # fmt: on
    ]



from typing import Any, Callable, TypeVar

import requests

from ..utils.console.logger import Logger
from .exceptions import MaxRetryError

T = TypeVar("T")


class MaxRetry:
    def __init__(self, max_retry: int = 2):
        self.max_retry = max_retry

    def __call__(self, connect_once: Callable[..., T]) -> Callable[..., T]:
        def connect_n_times(*args: Any, **kwargs: Any) -> T:
            retry = self.max_retry + 1
            while retry:
                try:
                    return connect_once(*args, **kwargs)
                except requests.exceptions.Timeout as e:
                    Logger.warning("抓取超时，正在尝试重新连接～")
                finally:
                    retry -= 1
            raise MaxRetryError()

        return connect_n_times



from typing import List
from enum import Enum


class Media(Enum):
    VIDEO = 0
    AUDIO = 30200


video_quality_sequence_default = [127, 125, 120, 116, 112, 80, 74, 64, 32, 16]
audio_quality_sequence_default = [30280, 30232, 30216]

video_quality_map = {
    127: {
        "description": "超高清 8K",
        "width": 8192,
        "height": 4320,
    },
    125: {
        "description": "HDR 真彩",
        "width": 3840,
        "height": 1920,
    },
    120: {
        "description": "4K 超清",
        "width": 3840,
        "height": 1920,
    },
    116: {
        "description": "1080P 60帧",
        "width": 2160,
        "height": 1080,
    },
    112: {
        "description": "1080P 高码率",
        "width": 2160,
        "height": 1080,
    },
    80: {
        "description": "1080P 高清",
        "width": 2160,
        "height": 1080,
    },
    74: {
        "description": "720P 60帧",
        "width": 1440,
        "height": 720,
    },
    64: {
        "description": "720P 高清",
        "width": 1440,
        "height": 720,
    },
    32: {
        "description": "480P 清晰",
        "width": 960,
        "height": 480,
    },
    16: {
        "description": "360P 流畅",
        "width": 720,
        "height": 360,
    },
    6: {
        "description": "240P 极速",
        "width": 320,
        "height": 240,
    },
    208: {
        "description": "1080P 高清",
        "width": 1920,
        "height": 1080,
    },
    192: {
        "description": "720P 高清",
        "width": 1280,
        "height": 720,
    },
}

audio_quality_map = {
    30280: {
        "description": "320kbps",
        "bitrate": 320,
    },
    30232: {
        "description": "128kbps",
        "bitrate": 128,
    },
    30216: {
        "description": "64kbps",
        "bitrate": 64,
    },
    0: {"description": "Unknown", "bitrate": 0},
}


def gen_quality_sequence(quality: int = 127, type: Media = Media.VIDEO) -> List[int]:
    """根据默认先降后升的清晰度机制生成清晰度序列"""
    quality_sequence_default = {
        Media.VIDEO: video_quality_sequence_default,
        Media.AUDIO: audio_quality_sequence_default,
    }[type]
    return quality_sequence_default[quality_sequence_default.index(quality) :] + list(
        reversed(quality_sequence_default[: quality_sequence_default.index(quality)])
    )



import pytest

from bilili.api.acg_video import (
    get_video_info,
    get_acg_video_title,
    get_acg_video_list,
    get_acg_video_playurl,
    get_acg_video_subtitle,
)


@pytest.mark.api
def test_get_video_info():
    bvid = "BV1vZ4y1M7mQ"
    assert get_video_info(bvid=bvid)["bvid"] == bvid


@pytest.mark.api
def test_get_title():
    bvid = "BV1vZ4y1M7mQ"
    assert get_acg_video_title(bvid=bvid) == "用 bilili 下载 B 站视频"


@pytest.mark.api
def test_get_list():
    bvid = "BV1vZ4y1M7mQ"
    video_list = get_acg_video_list(bvid=bvid)
    assert video_list[0]["cid"] == "222190584"
    assert video_list[1]["cid"] == "222200470"


@pytest.mark.api
@pytest.mark.ci_skip
@pytest.mark.parametrize("type", ["flv", "mp4", "dash"])
def test_get_playurl(type: str):
    bvid = "BV1vZ4y1M7mQ"
    cid = "222190584"
    play_list = get_acg_video_playurl(bvid=bvid, cid=cid, quality=120, audio_quality=30280, type=type)


@pytest.mark.api
def test_get_subtitle():
    bvid = "BV1i741187Dp"
    cid = "149439373"
    get_acg_video_subtitle(bvid=bvid, cid=cid)



import pytest
import subprocess
import sys

from bilili.__version__ import VERSION as bilili_version

PYTHON = sys.executable


@pytest.mark.e2e
def test_version_e2e():
    p = subprocess.run([PYTHON, "-m", "bilili", "-v"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)
    res = p.stdout.decode()
    assert res.strip().endswith(bilili_version)


@pytest.mark.e2e
def test_ui_e2e():
    p = subprocess.run(
        [PYTHON, "-m", "bilili.utils.console.ui"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True
    )


@pytest.mark.e2e
@pytest.mark.ci_skip
def test_bangumi_e2e():
    short_bangumi = "https://www.bilibili.com/bangumi/play/ep100367"
    p = subprocess.run(
        [PYTHON, "-m", "bilili", short_bangumi, "-p=^", "-q=16", "-y", "-w"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=True,
    )


@pytest.mark.e2e
def test_acg_video_e2e():
    short_acg_video = "https://www.bilibili.com/video/BV1AZ4y147Yg"
    p = subprocess.run(
        [PYTHON, "-m", "bilili", short_acg_video, "-q=16", "-y", "-w"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=True,
    )


@pytest.mark.e2e
def test_acg_video_8k_e2e():
    acg_video_8k = "https://www.bilibili.com/video/BV1qM4y1w716"
    p = subprocess.run(
        [PYTHON, "-m", "bilili", acg_video_8k, "-q=127", "-y", "-c=*****"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=True,
    )



import re

from .handlers.status import DownloaderStatus
from .utils.crawler import BililiCrawler
from .utils.functiontools.attrdict import AttrDict

# avid
regex_acg_video_av = re.compile(r"https?://(www\.|m\.)?bilibili\.com/video/av(?P<avid>\d+)")
regex_acg_video_av_short = re.compile(r"https?://b23\.tv/av(?P<avid>\d+)")

# bvid
regex_acg_video_bv = re.compile(r"https?://(www\.|m\.)?bilibili\.com/video/(?P<bvid>(bv|BV)\w+)")
regex_acg_video_bv_short = re.compile(r"https?://b23\.tv/(?P<bvid>(bv|BV)\w+)")

# media id
regex_bangumi_md = re.compile(r"https?://(www\.|m\.)?bilibili\.com/bangumi/media/md(?P<media_id>\d+)")

# episode id
regex_bangumi_ep = re.compile(r"https?://(www\.|m\.)?bilibili\.com/bangumi/play/ep(?P<episode_id>\d+)")
regex_bangumi_ep_short = re.compile(r"https?://b23\.tv/ep(?P<episode_id>\d+)")

# season id
regex_bangumi_ss = re.compile(r"https?://(www\.|m\.)?bilibili\.com/bangumi/play/ss(?P<season_id>\d+)")
regex_bangumi_ss_short = re.compile(r"https?://b23\.tv/ss(?P<season_id>\d+)")


spider = BililiCrawler()
global_status = DownloaderStatus()
# fmt: off
regex = {
    "acg_video": {
        "av": {
            "origin": regex_acg_video_av,
            "short": regex_acg_video_av_short,
        },
        "bv": {
            "origin": regex_acg_video_bv,
            "short": regex_acg_video_bv_short,
        },
    },
    "bangumi": {
        "md": {
            "origin": regex_bangumi_md,
        },
        "ep": {
            "origin": regex_bangumi_ep,
            "short": regex_bangumi_ep_short,
        },
        "ss": {
            "origin": regex_bangumi_ss,
            "short": regex_bangumi_ss_short,
        },
    },
} >> AttrDict()
# fmt: on



class APIException(Exception):
    def __init__(self, code, message):
        super().__init__(code, message)
        self.code = code
        self.message = message


class ArgumentsError(APIException):
    def __init__(self, *args):
        message = "参数 " + ",".join(args) + " 均空"
        super().__init__(101, message)


class CannotDownloadError(APIException):
    def __init__(self, code, message):
        message = "「{}」 {}".format(code, message)
        super().__init__(102, message)


class UnknownTypeError(APIException):
    def __init__(self, type):
        message = "未知类型：{}".format(type)
        super().__init__(103, message)


class UnsupportTypeError(APIException):
    def __init__(self, type):
        message = "不受支持的类型：{}".format(type)
        super().__init__(104, message)


class IsPreviewError(APIException):
    def __init__(self):
        message = "本视频是预览视频"
        super().__init__(105, message)


class MaxRetryError(APIException):
    def __init__(self):
        message = "超出最大重试次数"
        super().__init__(106, message)



import argparse
import os
import re
import sys
import time
from typing import List
from urllib.parse import quote, unquote

from .__version__ import VERSION as bilili_version
from .api.danmaku import get_danmaku
from .api.exceptions import CannotDownloadError, IsPreviewError
from .api.vip import is_vip
from .handlers.downloader import RemoteFile
from .handlers.merger import MergingFile
from .tools import global_status, regex, spider
from .utils.base import repair_filename, size_format, touch_dir
from .utils.console.colorful import set_no_color
from .utils.console.logger import Badge, Logger, set_logger_debug
from .utils.console.ui import ColorString, Line, LineList, ProgressBar, String, View
from .utils.danmaku import convert_xml_danmaku_to_ass
from .utils.functiontools.attrdict import AttrDict
from .utils.playlist import Dpl, M3u
from .utils.subtitle import Subtitle
from .utils.thread import Flag, ThreadPool
from .video import BililiContainer


def parse_episodes(episodes_str: str, total: int) -> List[int]:
    """将选集字符串转为列表"""

    if total == 0:
        Logger.warning("该剧集列表无任何剧集，猜测正片尚未上线，如果想要下载 PV 等特殊剧集，请添加参数 -s")
        return []

    def reslove_negetive(value: int) -> int:
        return value if value > 0 else value + total + 1

    # 解析字符串为列表
    Logger.print("全 {} 话".format(total))
    if re.match(r"([\-\d\^\$]+(~[\-\d\^\$]+)?)(,[\-\d\^\$]+(~[\-\d\^\$]+)?)*", episodes_str):
        episodes_str = episodes_str.replace("^", "1")
        episodes_str = episodes_str.replace("$", "-1")
        episode_list = []
        for episode_item in episodes_str.split(","):
            if "~" in episode_item:
                start, end = episode_item.split("~")
                start, end = int(start), int(end)
                start, end = reslove_negetive(start), reslove_negetive(end)
                assert end >= start, "终点值（{}）应不小于起点值（{}）".format(end, start)
                episode_list.extend(list(range(start, end + 1)))
            else:
                episode_item = int(episode_item)
                episode_item = reslove_negetive(episode_item)
                episode_list.append(episode_item)
    else:
        episode_list = []

    episode_list = sorted(list(set(episode_list)))

    # 筛选满足条件的剧集
    out_of_range = []
    episodes = []
    for episode in episode_list:
        if episode in range(1, total + 1):
            if episode not in episodes:
                episodes.append(episode)
        else:
            out_of_range.append(episode)
    if out_of_range:
        Logger.warning("剧集 {} 不存在哟！".format(",".join(list(map(str, out_of_range)))))

    Logger.print("已选择第 {} 话".format(",".join(list(map(str, episodes)))))
    assert episodes, "没有选中任何剧集"
    return episodes


def check_arguments_and_set_global(args: argparse.Namespace):
    # 先解码后编码是防止获取到的 SESSDATA 是已经解码后的（包含「,」）
    # 而番剧无法使用解码后的 SESSDATA
    cookies = {"SESSDATA": quote(unquote(args.sess_data))}
    spider.set_cookies(cookies)

    if args.debug:
        set_logger_debug()

    # 使用 --no-color 或者 NO_COLOR 环境变量非空均不显示颜色
    if args.no_color or os.environ.get("NO_COLOR") is not None:
        set_no_color()

    if args.disable_proxy:
        spider.trust_env = False

    # 大会员身份校验
    if not args.sess_data:
        Logger.info("未提供 SESSDATA，无法下载会员专享剧集的喔～")
    else:
        if is_vip():
            Logger.custom("成功以大会员身份登录～", badge=Badge("大会员", fore="white", back="magenta", style="bold"))
        else:
            Logger.warning("以非大会员身份登录，无法下载会员专享剧集的喔～")


def main():
    """解析命令行参数并调用相关模块进行下载"""

    parser = argparse.ArgumentParser(description="bilili B 站视频、弹幕下载器", prog="bilili")
    parser.add_argument("-v", "--version", action="version", version="%(prog)s {}".format(bilili_version))
    parser.add_argument("url", help="视频主页地址")
    parser.add_argument(
        "-t",
        "--type",
        default="dash",
        choices=["flv", "dash", "mp4"],
        help="选择下载源类型（dash 或 flv 或 mp4）",
    )
    parser.add_argument("-d", "--dir", default=r"", help="下载目录")
    parser.add_argument(
        "-q",
        "--quality",
        default=127,
        choices=[127, 125, 120, 116, 112, 80, 74, 64, 32, 16],
        type=int,
        help="视频清晰度 127:8K, 125:HDR, 120:4K, 116:1080P60, 112:1080P+, 80:1080P, 74:720P60, 64:720P, 32:480P, 16:360P",
    )
    parser.add_argument("-n", "--num-threads", default=16, type=int, help="最大下载线程数")
    parser.add_argument("-p", "--episodes", default="^~$", help="选集")
    parser.add_argument("-s", "--with-section", action="store_true", help="同时下载附加剧集（PV、预告以及特别篇等专区内容）")
    parser.add_argument("-w", "--overwrite", action="store_true", help="强制覆盖已下载视频")
    parser.add_argument("-c", "--sess-data", default="", help="输入 cookies")
    parser.add_argument("-y", "--yes", action="store_true", help="跳过下载询问")
    parser.add_argument(
        "--audio-quality",
        default=30280,
        choices=[30280, 30232, 30216],
        type=int,
        help="音频码率等级 30280:320kbps, 30232:128kbps, 30216:64kbps",
    )
    parser.add_argument(
        "--playlist-type",
        default="dpl",
        choices=["dpl", "m3u", "no"],
        help="播放列表类型，支持 dpl 和 m3u，输入 no 不生成播放列表",
    )
    parser.add_argument(
        "--danmaku",
        default="xml",
        choices=["xml", "ass", "no"],
        help="弹幕类型，支持 xml 和 ass，如果设置为 no 则不下载弹幕",
    )
    parser.add_argument(
        "--block-size",
        default=128,
        type=int,
        help="分块下载器的块大小，单位为 MiB，默认为 128MiB，设置为 0 时禁用分块下载",
    )
    parser.add_argument("--abs-path", action="store_true", help="修改播放列表路径类型为绝对路径")
    parser.add_argument("--use-mirrors", action="store_true", help="启用从多个镜像下载功能")
    parser.add_argument("--disable-proxy", action="store_true", help="禁用系统代理")
    parser.add_argument("--no-color", action="store_true", help="不使用颜色")
    parser.add_argument("--debug", action="store_true", help="debug 模式")

    args = parser.parse_args()
    check_arguments_and_set_global(args)

    config = {
        "url": args.url,
        "dir": args.dir,
        "quality": args.quality,
        "audio_quality": args.audio_quality,
        "with_section": args.with_section,
        "episodes": args.episodes,
        "playlist_type": args.playlist_type,
        "playlist_path_type": "AP" if args.abs_path else "RP",
        "overwrite": args.overwrite,
        "type": args.type.lower(),
        "block_size": int(args.block_size * 1024 * 1024),
    } >> AttrDict()

    # 匹配资源的 id 以及其对应所属类型
    # fmt: off
    resource_id = {
        "avid": "",
        "bvid": "",
        "episode_id": "",
        "season_id": "",
    } >> AttrDict()
    # fmt: on

    # fmt: off
    if (avid_match := regex.acg_video.av.origin.match(args.url)) or \
        (avid_match := regex.acg_video.av.short.match(args.url)):
        from .api.acg_video import get_video_info
        avid = avid_match.group("avid")
        if episode_id := get_video_info(avid=avid)["episode_id"]:
            resource_id.episode_id = episode_id
        else:
            resource_id.avid = avid
    elif (bvid_match := regex.acg_video.bv.origin.match(args.url)) or \
        (bvid_match := regex.acg_video.bv.short.match(args.url)):
        from .api.acg_video import get_video_info
        bvid = bvid_match.group("bvid")
        if episode_id := get_video_info(bvid=bvid)["episode_id"]:
            resource_id.episode_id = episode_id
        else:
            resource_id.bvid = bvid
    elif media_id_match := regex.bangumi.md.origin.match(args.url):
        from .api.bangumi import get_season_id
        media_id = media_id_match.group("media_id")
        resource_id.season_id = get_season_id(media_id=media_id)
    elif (episode_id_match := regex.bangumi.ep.origin.match(args.url)) or \
        (episode_id_match := regex.bangumi.ep.short.match(args.url)):
        episode_id = episode_id_match.group("episode_id")
        resource_id.episode_id = episode_id
    elif (season_id_match := regex.bangumi.ss.origin.match(args.url)) or \
        (season_id_match := regex.bangumi.ss.short.match(args.url)):
        season_id = season_id_match.group("season_id")
        resource_id.season_id = season_id
    else:
        Logger.error("视频地址有误呀，请仔细检查一下下～")
        sys.exit(1)
    # fmt: on

    if resource_id.avid or resource_id.bvid:
        from .parser.acg_video import get_list, get_playurl, get_subtitle, get_title

        bili_type = "acg_video"
    elif resource_id.season_id or resource_id.episode_id:
        from .parser.bangumi import get_list, get_playurl, get_subtitle, get_title

        bili_type = "bangumi"
    else:
        Logger.error("未知的视频类型！")
        sys.exit(1)

    # 获取标题
    title = get_title(resource_id)
    Logger.print(title)

    # 创建所需目录结构
    base_dir = touch_dir(os.path.join(config["dir"], repair_filename(title + " - bilibili")))
    video_dir = touch_dir(os.path.join(base_dir, "Videos"))

    # 获取需要的信息
    containers = [
        BililiContainer(video_dir=video_dir, type=args.type, **video)
        for video in get_list(resource_id, config["with_section"])
    ]

    # 解析并过滤不需要的选集
    episodes = parse_episodes(config["episodes"], len(containers))
    containers, containers_need_filter = [], containers
    for container in containers_need_filter:
        if container.id not in episodes:
            container._.downloaded = True
            container._.merged = True
        else:
            containers.append(container)

    # 初始化播放列表
    if config["playlist_type"] == "dpl":
        playlist = Dpl(os.path.join(base_dir, "Playlist.dpl"), path_type=config["playlist_path_type"])
    elif config["playlist_type"] == "m3u":
        playlist = M3u(os.path.join(base_dir, "Playlist.m3u"), path_type=config["playlist_path_type"])
    else:
        playlist = None

    # 解析片段信息及视频 url
    for i, container in enumerate(containers):
        Logger.print(
            "{:02}/{:02} 正在努力解析视频信息～".format(i + 1, len(containers)),
            end="\r",
        )

        # 解析视频 url
        try:
            for playinfo in get_playurl(container, config["quality"], config["audio_quality"]):
                container.append_media(block_size=config["block_size"], **playinfo)
        except CannotDownloadError as e:
            Logger.warning("{} 无法下载，原因：{}".format(container.name, e.message))
            del containers[i]
            continue
        except IsPreviewError:
            # TODO: 现在还有部分预览的视频吗？
            Logger.warning("{} 是预览视频呢～".format(container.name))

        # 写入播放列表
        if playlist is not None:
            playlist.write_path(container.path)

        # 下载字幕
        for sub_info in get_subtitle(container):
            sub_path = "{}_{}.srt".format(os.path.splitext(container.path)[0], sub_info["lang"])
            subtitle = Subtitle(sub_path)
            for sub_line in sub_info["lines"]:
                subtitle.write_line(sub_line["content"], sub_line["from"], sub_line["to"])

        # 生成弹幕
        if args.danmaku != "no":
            xml_danmaku = get_danmaku(container.meta["cid"])
            if args.danmaku == "ass":
                with open(
                    os.path.splitext(container.path)[0] + ".ass",
                    "w",
                    encoding="utf-8-sig",
                    errors="replace",
                ) as f:
                    f.write(
                        convert_xml_danmaku_to_ass(
                            xml_danmaku,
                            container.height,
                            container.width,
                        )
                    )
            else:
                with open(os.path.splitext(container.path)[0] + ".xml", "w", encoding="utf-8") as f:
                    f.write(xml_danmaku)

    if playlist is not None:
        playlist.flush()

    # 准备下载
    if containers:
        # 状态检查与校正
        for i, container in enumerate(containers, 1):
            container_downloaded = not container.check_needs_download(args.overwrite)
            symbol = " " if container_downloaded else "*"
            if container_downloaded:
                container._.merged = True
            Logger.print("{}{} {:>2} {}".format("    " * 0, symbol, i, str(container)))
            for media in container.medias:
                media_downloaded = not media.check_needs_download(args.overwrite) or container_downloaded
                symbol = " " if media_downloaded else "*"
                if not container_downloaded and args.debug:
                    Logger.print("{}{} {}".format("    " * 1, symbol, media.name))
                for block in media.blocks:
                    block_downloaded = not block.check_needs_download(args.overwrite) or media_downloaded
                    symbol = " " if block_downloaded else "*"
                    block._.downloaded = block_downloaded
                    if not media_downloaded and args.debug:
                        Logger.print("{}{} {}".format("    " * 2, symbol, block.name))

        # 询问是否下载，通过参数 -y 可以跳过
        if not args.yes:
            answer = None
            while answer is None:
                result = input("以上标 * 为需要进行下载的视频，是否立刻进行下载？[Y/n]")
                if result == "" or result[0].lower() == "y":
                    answer = True
                elif result[0].lower() == "n":
                    answer = False
                else:
                    answer = None
            if not answer:
                sys.exit(0)

        # 部署下载与合并任务
        merge_wait_flag = Flag(False)  # 合并线程池不能因为没有任务就结束
        # 因此要设定一个 flag，待最后合并结束后改变其值
        merge_pool = ThreadPool(3, wait=merge_wait_flag, daemon=True)
        download_pool = ThreadPool(
            args.num_threads,
            daemon=True,
            thread_globals_creator={
                # 为每个线程创建一个全新的 Session，因为 requests.Session 不是线程安全的
                # https://github.com/psf/requests/issues/1871
                "thread_spider": spider.clone
            },
        )
        for container in containers:
            merging_file = MergingFile(
                container.type,
                [media.path for media in container.medias],
                container.path,
            )
            for media in container.medias:

                block_merging_file = MergingFile(None, [block.path for block in media.blocks], media.path)
                for block in media.blocks:

                    mirrors = block.mirrors if args.use_mirrors else []
                    remote_file = RemoteFile(block.url, block.path, mirrors=mirrors, range=block.range)

                    # 为下载挂载各种钩子，以修改状态，注意外部变量应当作为默认参数传入
                    @remote_file.on("before_download")
                    def before_download(file, status=block._):
                        status.downloading = True

                    @remote_file.on("updated")
                    def updated(file, status=block._):
                        status.size = file.size

                    @remote_file.on("downloaded")
                    def downloaded(
                        file, status=block._, merging_file=merging_file, block_merging_file=block_merging_file
                    ):
                        status.downloaded = True

                        if status.parent.downloaded:
                            # 当前 media 的最后一个 block 所在线程进行合并（直接执行，不放线程池）
                            status.downloaded = False
                            block_merging_file.merge()
                            status.downloaded = True

                            # 如果该线程同时也是当前 container 的最后一个 block，就部署合并任务（放到线程池）
                            if status.parent.parent.downloaded and not status.parent.parent.merged:
                                # 为合并挂载各种钩子
                                @merging_file.on("before_merge")
                                def before_merge(file, status=status.parent.parent):
                                    status.merging = True

                                @merging_file.on("merged")
                                def merged(file, status=status.parent.parent):
                                    status.merging = False
                                    status.merged = True

                                merge_pool.add_task(merging_file.merge, args=())

                        status.downloading = False

                    # 下载过的不应继续部署任务
                    if block._.downloaded:
                        continue
                    download_pool.add_task(remote_file.download, args=())

        # 启动线程池
        merge_pool.run()
        download_pool.run()

        # 初始化界面
        console = View(debug=args.debug)
        console.add_component(Line(center=String(), fillchar=" "))
        console.add_component(Line(left=ColorString(fore="cyan"), fillchar=" "))
        console.add_component(LineList(Line(left=String(), right=String(), fillchar="-")))
        console.add_component(
            Line(
                left=ColorString(
                    fore="green",
                    back="white",
                    subcomponent=ProgressBar(symbols=" ▏▎▍▌▋▊▉█", width=65),
                ),
                right=String(),
                fillchar=" ",
            )
        )
        console.add_component(Line(left=ColorString(fore="blue"), fillchar=" "))
        console.add_component(LineList(Line(left=String(), fillchar=" ")))
        console.add_component(
            Line(
                left=ColorString(
                    fore="yellow",
                    back="white",
                    subcomponent=ProgressBar(symbols=" ▏▎▍▌▋▊▉█", width=65),
                ),
                right=String(),
                fillchar=" ",
            )
        )

        # 准备监控
        size, t = global_status.size, time.time()
        while True:
            now_size, now_t = global_status.size, time.time()
            delta_size, delta_t = (
                max(now_size - size, 0),
                (now_t - t) if now_t - t > 1e-6 else 1e-6,
            )
            speed = delta_size / delta_t
            size, t = now_size, now_t

            # 数据传入，界面渲染
            console.refresh(
                # fmt: off
                [
                    {
                        "center": " bilili ",
                    },
                    {
                        "left": "Downloading videos: "
                    } if global_status.downloading else None,
                    [
                        {
                            "left": "{} ".format(str(container)),
                            "right": " {}/{}".format(
                                size_format(container._.size), size_format(container._.total_size),
                            ),
                        } if container._.downloading else None
                        for container in containers
                    ] if global_status.downloading else None,
                    {
                        "left": global_status.size / global_status.total_size,
                        "right": " {}/{} {}/s".format(
                            size_format(global_status.size),
                            size_format(global_status.total_size),
                            size_format(speed),
                        ),
                    } if global_status.downloading else None,
                    {
                        "left": "Merging videos: "
                    } if global_status.merging else None,
                    [
                        {
                            "left": "{} ".format(str(container)),
                            "right": True
                        } if container._.merging else None
                        for container in containers
                    ] if global_status.merging else None,
                    {
                        "left": sum([container._.merged for container in containers]) / len(containers),
                        "right": " {}/{}".format(
                            sum([container._.merged for container in containers]), len(containers),
                        ),
                    } if global_status.merging else None,
                ]
                # fmt: on
            )

            # 检查是否已经全部完成
            if global_status.downloaded and global_status.merged:
                merge_wait_flag.value = True
                download_pool.join()
                merge_pool.join()
                break
            try:
                # 将刷新率稳定在 2fps
                refresh_rate = 2
                time.sleep(max(1 / refresh_rate - (time.time() - now_t), 0.01))
            except (SystemExit, KeyboardInterrupt):
                Logger.info("已终止下载，再次运行即可继续下载～")
                sys.exit(1)
        Logger.info("已全部下载完成啦！")
    else:
        Logger.info("没有需要下载的视频！")


if __name__ == "__main__":
    main()



from biliass import Danmaku2ASS


def convert_xml_danmaku_to_ass(xml_text: str, height: int, width: int) -> str:
    return Danmaku2ASS(
        xml_text,
        width,
        height,
        input_format="xml",
        reserve_blank=0,
        font_face="sans-serif",
        font_size=width / 40,
        text_opacity=0.8,
        duration_marquee=15.0,
        duration_still=10.0,
        comment_filter=None,
        is_reduce_comments=False,
        progress_callback=None,
    )



import os
import random
import re
from typing import Any, Tuple, Optional


class Ref:
    """引用类

    用于包裹基本数据类型，将其封装为对象，其值通过 var.value 来访问
    """

    def __init__(self, value: Any):
        self.value = value


class Writer:
    """ 文件写入器，持续打开文件对象，直到使用完毕后才关闭 """

    def __init__(self, path: str, mode: str = "wb", **kwargs: Any):
        self.path = path
        self._f = open(path, mode, **kwargs)

    def __del__(self):
        self._f.close()

    def flush(self):
        self._f.flush()

    def write(self, content: str):
        self._f.write(content)


class Text(Writer):
    """ 文本写入器 """

    def __init__(self, path: str, **kwargs: Any):
        kwargs["encoding"] = kwargs.get("encoding", "utf-8")
        super().__init__(path, "w", **kwargs)

    def write_string(self, string: str):
        self.write(string + "\n")


def touch_dir(path: str) -> str:
    """ 若文件夹不存在则新建，并返回标准路径 """
    if not os.path.exists(path):
        os.makedirs(path)
    return os.path.normpath(path)


def touch_file(path: str) -> str:
    """ 若文件不存在则新建，并返回标准路径 """
    if not os.path.exists(path):
        open(path, "w").close()
    return os.path.normpath(path)


def touch_url(url: str, spider) -> Tuple[Optional[str], bool]:
    """ 与资源进行测试连接，并获取该资源的 size 与 是否可以断点续传 """
    # 某些资源 head 无法获得真实 size
    methods = [spider.head, spider.get]
    for method in methods:
        res = method(url, headers={"Range": "bytes=0-4"})
        size, resumable = None, False
        if res.headers.get("Content-Range"):
            size = int(res.headers["Content-Range"].split("/")[-1])
            resumable = True
        elif res.headers.get("Content-Length"):
            size = int(res.headers["Content-Length"])
            resumable = False
        else:
            size = None
            resumable = False
        if size and resumable:
            break
    return size, resumable


def repair_filename(filename: str) -> str:
    """ 修复不合法的文件名 """

    def to_full_width_chr(matchobj: "re.Match[str]") -> str:
        char = matchobj.group(0)
        full_width_char = chr(ord(char) + ord("？") - ord("?"))
        return full_width_char

    # 路径非法字符，转全角
    regex_path = re.compile(r'[\\/:*?"<>|]')
    # 空格类字符，转空格
    regex_spaces = re.compile(r"\s+")
    # 不可打印字符，移除
    regex_non_printable = re.compile(
        r"[\001\002\003\004\005\006\007\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f"
        r"\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a]"
    )

    filename = regex_path.sub(to_full_width_chr, filename)
    filename = regex_spaces.sub(" ", filename)
    filename = regex_non_printable.sub("", filename)
    filename = filename.strip()
    filename = filename if filename else "file_{:04}".format(random.randint(0, 9999))
    return filename


def get_size(path: str) -> int:
    """ 获取文件夹或文件的字节数 """
    if os.path.isfile(path):
        return os.path.getsize(path)
    elif os.path.isdir(path):
        size = 0
        for subpath in os.listdir(path):
            size += get_size(os.path.join(path, subpath))
        return size
    else:
        return 0


def size_format(size: float, ndigits: int = 2) -> str:
    """ 输入数据字节数，与保留小数位数，返回数据量字符串 """
    flag = "-" if size < 0 else ""
    size = abs(size)
    units = ["Bytes", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB", "BiB"]
    idx = len(units) - 1
    unit = ""
    unit_size = 0
    while idx >= 0:
        unit_size = 2 ** (idx * 10)
        if size >= unit_size:
            unit = units[idx]
            break
        idx -= 1
    return "{}{:.{}f} {}".format(flag, size / unit_size, ndigits, unit)


def get_char_width(char: str) -> int:
    """ 计算单个字符的宽度 """
    # fmt: off
    widths = [
        (126, 1), (159, 0), (687, 1), (710, 0), (711, 1),
        (727, 0), (733, 1), (879, 0), (1154, 1), (1161, 0),
        (4347, 1), (4447, 2), (7467, 1), (7521, 0), (8369, 1),
        (8426, 0), (9000, 1), (9002, 2), (11021, 1), (12350, 2),
        (12351, 1), (12438, 2), (12442, 0), (19893, 2), (19967, 1),
        (55203, 2), (63743, 1), (64106, 2), (65039, 1), (65059, 0),
        (65131, 2), (65279, 1), (65376, 2), (65500, 1), (65510, 2),
        (120831, 1), (262141, 2), (1114109, 1),
    ]
    # fmt: on

    o = ord(char)
    if o == 0xE or o == 0xF:
        return 0
    for num, wid in widths:
        if o <= num:
            return wid
    return 1


def get_string_width(string: str) -> int:
    """ 计算包含中文的字符串宽度 """
    # 去除颜色码
    regex_color = re.compile(r"\033\[\d+m")
    string = regex_color.sub("", string)
    try:
        length = sum([get_char_width(c) for c in string])
    except:
        length = len(string)
    return length



import math
import os
import platform
import shutil
import sys
from typing import Any, List, Optional, Tuple, Union

from ..base import get_string_width
from .colorful import Back, Fore, Style, colored_string
from .logger import Logger

IS_WINDOWS = platform.system() == "Windows"


def get_terminal_size() -> Tuple[int, int]:
    """Get the size of the console.
    @refs: https://github.com/willmcgugan/rich/blob/e5246436cd75de32f3436cc88d6e4fdebe13bd8d/rich/console.py#L918-L951
    Returns:
        tuple[int, int]: A named tuple containing the dimensions.
    """

    width: Optional[int] = None
    height: Optional[int] = None
    if IS_WINDOWS:  # pragma: no cover
        width, height = shutil.get_terminal_size()
    else:
        try:
            width, height = os.get_terminal_size(sys.stdin.fileno())
        except (AttributeError, ValueError, OSError):
            try:
                width, height = os.get_terminal_size(sys.stdout.fileno())
            except (AttributeError, ValueError, OSError):
                pass

    width = width or 80
    height = height or 25
    return (width, height)


class View:
    max_width = 100
    min_width = 50

    def __init__(self, debug: bool = False):
        self.debug = debug
        self.components = []
        self.rendered_area = (0, 0)

    def add_component(self, component: "Component"):
        self.components.append(component)

    def render(self, data: Any) -> str:
        if data is None:
            return ""
        assert len(self.components) == len(data), "数据个数与组件个数不匹配"
        result = ""
        for component, component_data in zip(self.components, data):
            result += component.render(component_data)
        self.rendered_area = View.calc_area_size(result)
        return result

    def refresh(self, data: Any):
        if not self.debug:
            self.clear()
        Logger.print(self.render(data))

    def clear(self):
        os.system("cls" if os.name == "nt" else "clear")
        # 暂时无法判断什么终端支持什么终端不支持（如 cmd），暂时不使用
        # if self.rendered_area == (0, 0):
        #     return
        # clear_str = ""
        # clear_str += "\x1b[1A" * self.rendered_area[0]
        # clear_str += "\r"
        # # 这里不应该使用 Logger 打印，因为可能被去除颜色控制符（虽然现在不会）
        # print(clear_str, end="")

    @classmethod
    def get_width(cls) -> int:
        width = get_terminal_size()[0]
        width = min(width, View.max_width)
        width = max(width, View.min_width)
        return width

    @classmethod
    def get_height(cls) -> int:
        return get_terminal_size()[1]

    @classmethod
    def calc_area_size(cls, string: str) -> Tuple[int, int]:
        lines = string.split("\n")
        return (len(lines), get_string_width(lines[0]))


class Component:
    def __init__(self):
        pass

    def render(self, data: Any) -> str:
        raise NotImplementedError


class String(Component):
    def __init__(self):
        super().__init__()

    def render(self, data: Any) -> str:
        if data is None:
            return ""
        return data


class EndLine(Component):
    def __init__(self):
        super().__init__()

    def render(self, data: Any) -> str:
        if data is None:
            return ""
        return "\n"


class Font(Component):
    def __init__(self, char_a: str = "ａ", char_A: Optional[str] = None):
        super().__init__()
        self.char_a = char_a
        self.char_A = char_A

    def render(self, data: Any) -> str:
        if data is None:
            return ""
        result = ""
        for char in data:
            if ord(char) >= ord("a") and ord(char) <= ord("z"):
                result += chr(ord(char) + ord(self.char_a) - ord("a"))
            elif ord(char) >= ord("A") and ord(char) <= ord("Z"):
                if self.char_A is None:
                    result += chr(ord(char) + ord(self.char_a) - ord("a"))
                else:
                    result += chr(ord(char) + ord(self.char_A) - ord("A"))
            else:
                result += char
        return result


class ColorString(Component):
    def __init__(
        self,
        fore: Optional[Fore] = None,
        back: Optional[Back] = None,
        style: Optional[Style] = None,
        subcomponent: Optional[Component] = None,
    ):
        super().__init__()
        self.fore: Optional[Fore] = fore
        self.back: Optional[Back] = back
        self.style: Optional[Style] = style
        self.subcomponent: Optional[Component] = subcomponent

    def render(self, data: Any) -> str:
        if data is None:
            return ""
        subcomponet_string = self.subcomponent.render(data) if self.subcomponent is not None else data
        return colored_string(subcomponet_string, self.fore, self.back, self.style)


class Line(Component):
    def __init__(
        self,
        left: Optional[Component] = None,
        center: Optional[Component] = None,
        right: Optional[Component] = None,
        fillchar: str = " ",
    ):
        super().__init__()
        self.left = left
        self.center = center
        self.right = right
        self.fillchar = fillchar

    def render(self, data: Any) -> str:
        if data is None:
            return ""
        left_data = data.get("left", None)
        center_data = data.get("center", None)
        right_data = data.get("right", None)

        left_result, center_result, right_result = "", "", ""
        left_width, center_width, right_width = 0, 0, 0
        left_placeholder_width, right_placeholder_width = 0, 0
        if self.left is not None:
            assert left_data is not None
            left_result = self.left.render(left_data)
            left_width: int = get_string_width(left_result)
        if self.right is not None:
            assert right_data is not None
            right_result = self.right.render(right_data)
            right_width: int = get_string_width(right_result)
        if self.center is not None:
            assert center_data is not None
            center_result = self.center.render(center_data)
            center_width: int = get_string_width(center_result)

        if self.center is not None:
            left_placeholder_width = (View.get_width() - center_width) // 2 - left_width
            right_placeholder_width = (
                View.get_width() - left_width - left_placeholder_width - center_width - right_width
            )

            return (
                left_result
                + left_placeholder_width * self.fillchar
                + center_result
                + right_placeholder_width * self.fillchar
                + right_result
                + "\n"
            )
        else:
            left_placeholder_width = View.get_width() - left_width - right_width
            return left_result + left_placeholder_width * self.fillchar + right_result + "\n"


class Center(Component):
    def __init__(self, fillchar: str = " "):
        super().__init__()
        self.fillchar = fillchar

    def render(self, data: Any) -> str:
        if data is None:
            return ""
        return data.center(View.get_width(), self.fillchar) + "\n"


class ProgressBar(Component):
    def __init__(self, symbols: Union[str, List[str]] = "░▏▎▍▌▋▊▉█", width: int = View.get_width()):
        super().__init__()
        self.width = width
        self.symbols = symbols
        assert len(symbols) >= 2, "symbols 至少为 2 个"
        self.num_symbol = len(symbols)

    def render(self, data: Any) -> str:
        if data is None:
            return ""
        if data == 1:
            return self.symbols[-1] * self.width
        length = self.width * data
        length_int = int(length)
        length_float = length - length_int

        return (
            length_int * self.symbols[-1]
            + self.symbols[math.floor(length_float * (self.num_symbol - 1))]
            + (self.width - length_int - 1) * self.symbols[0]
        )


class DynamicSymbol(Component):
    def __init__(self, symbols: Union[str, List[str]] = "⠁⠂⠄⡀⢀⠠⠐⠈"):
        super().__init__()
        self.symbols = symbols
        self.index = 0

    def render(self, data: Any) -> str:
        if data is None:
            return ""
        self.index += 1
        self.index %= len(self.symbols)
        return self.symbols[self.index]


class LineList(Component):
    def __init__(self, subcomponent: Component):
        super().__init__()
        self.subcomponent = subcomponent

    def render(self, data: Any) -> str:
        if data is None:
            return ""
        result = ""
        for item in data:
            result += self.subcomponent.render(item)
        return result


if __name__ == "__main__":
    import time

    console = View()
    console.add_component(Line(center=Font(char_a="𝓪", char_A="𝓐"), fillchar="="))
    console.add_component(Line(left=ColorString(fore="cyan", style="italic"), fillchar=" "))
    console.add_component(LineList(Line(left=String(), right=String(), fillchar="-")))
    console.add_component(Line(left=ColorString(fore="blue", style="italic"), fillchar=" "))
    console.add_component(LineList(Line(left=String(), right=String(), fillchar="-")))
    console.add_component(
        Line(
            left=ColorString(
                fore="green",
                back="white",
                subcomponent=ProgressBar(symbols=" ▏▎▍▌▋▊▉█", width=70),
            ),
            right=String(),
            fillchar=" ",
        )
    )
    for i in range(100):
        # fmt: off
        console.refresh([
            {
                'center': ' 🍻 bilili ',
            },
            {
                'left': '🌠 Downloading videos:'
            },
            [
                {'left': '视频 1 ', 'right': ' 50%'},
                {'left': '视频 2 ', 'right': ' 40%'}
            ],
            {
                'left': '🍰 Merging videos:'
            },
            [
                {'left': '视频 3 ', 'right': ' 50%'},
                {'left': '视频 4 ', 'right': ' 40%'}
            ],
            {
                'left': (i+1) / 100,
                'right': "100MB/123MB 11.2 MB/s ⚡"
            }
        ])
        # fmt: on
        time.sleep(0.01)



import os
from typing import Callable, List

from ..handlers.base import Handler
from ..utils.console.logger import Logger
from ..utils.ffmpeg import FFmpeg

ffmpeg = FFmpeg()


class MergingFile(Handler):
    before_merge: Callable[..., None]
    merged: Callable[..., None]

    def __init__(self, type: str, src_path_list: List[str] = [], dst_path: str = ""):
        super().__init__(["before_merge", "merged"])
        self.type = type
        self.src_path_list = src_path_list
        self.dst_path = dst_path

    def merge(self):
        self.before_merge(self)
        if self.type == "mp4" or self.type is None:
            with open(self.dst_path, "wb") as fw:
                for src_path in self.src_path_list:
                    with open(src_path, "rb") as fr:
                        fw.write(fr.read())
        elif self.type == "flv":
            ffmpeg.join_videos(self.src_path_list, self.dst_path)
        elif self.type == "dash":
            if len(self.src_path_list) == 2:
                ffmpeg.join_video_audio(self.src_path_list[0], self.src_path_list[1], self.dst_path)
            else:
                ffmpeg.convert(self.src_path_list[0], self.dst_path)
        else:
            Logger.error("未知类型： {}".format(self.type))
        for src_path in self.src_path_list:
            os.remove(src_path)
        self.merged(self)



from ..api.bangumi import get_bangumi_list, get_bangumi_playurl, get_bangumi_subtitle, get_bangumi_title


def get_title(resource_id):
    return get_bangumi_title(season_id=resource_id.season_id, episode_id=resource_id.episode_id)


def get_list(resource_id, with_section: bool = False):
    video_list = get_bangumi_list(
        season_id=resource_id.season_id, episode_id=resource_id.episode_id, with_section=with_section
    )
    return [
        # fmt: off
        {
            "id": video["id"],
            "name": video["name"],
            "meta": {
                "avid": video["avid"],
                "bvid": video["bvid"],
                "cid": video["cid"],
                "episode_id": video["episode_id"]
            },
        }
        # fmt: on
        for video in video_list
    ]


def get_playurl(container, quality, audio_quality):
    play_list = get_bangumi_playurl(
        avid=container.meta["avid"],
        episode_id=container.meta["episode_id"],
        cid=container.meta["cid"],
        quality=quality,
        audio_quality=audio_quality,
        type=container.type,
    )
    return [
        {
            "id": play_info["id"],
            "url": play_info["url"],
            "mirrors": play_info["mirrors"],
            "quality": play_info["quality"],
            "height": play_info["height"],
            "width": play_info["width"],
            "size": play_info["size"],
            "type": play_info["type"],
        }
        for play_info in play_list
    ]


def get_subtitle(container):
    return get_bangumi_subtitle(
        avid=container.meta["avid"],
        bvid=container.meta["bvid"],
        cid=container.meta["cid"],
    )



import json
import re

from ..api.exceptions import ArgumentsError, CannotDownloadError, UnknownTypeError, UnsupportTypeError
from ..quality import Media, gen_quality_sequence, video_quality_map
from ..tools import regex_bangumi_ep, spider
from ..utils.base import touch_url
from .utils import MaxRetry


@MaxRetry(2)
def get_video_info(avid: str = "", bvid: str = ""):
    if not (avid or bvid):
        raise ArgumentsError("avid", "bvid")
    info_api = "http://api.bilibili.com/x/web-interface/view?aid={avid}&bvid={bvid}"
    res = spider.get(info_api.format(avid=avid, bvid=bvid), timeout=3)
    res_json_data = res.json()["data"]
    episode_id = ""
    if res_json_data.get("redirect_url") and (match_obj := regex_bangumi_ep.match(res_json_data["redirect_url"])):
        episode_id = match_obj.group("episode_id")
    return {
        "avid": str(res_json_data["aid"]),
        "bvid": res_json_data["bvid"],
        "title": res_json_data["title"],
        "picture": res_json_data["pic"],
        "episode_id": episode_id,
    }


def get_acg_video_title(avid: str = "", bvid: str = "") -> str:
    if not (avid or bvid):
        raise ArgumentsError("avid", "bvid")
    try:
        title = get_video_info(avid=avid, bvid=bvid)["title"]
    except:
        title = "呐，我也不知道是什么标题呢～"
    finally:
        return title


@MaxRetry(2)
def get_acg_video_list(avid: str = "", bvid: str = ""):
    if not (avid or bvid):
        raise ArgumentsError("avid", "bvid")
    list_api = "https://api.bilibili.com/x/player/pagelist?aid={avid}&bvid={bvid}&jsonp=jsonp"
    res = spider.get(list_api.format(avid=avid, bvid=bvid), timeout=3)
    return [
        # fmt: off
        {
            'id': i + 1,
            'name': item['part'],
            'cid': str(item['cid'])
        }
        for i, item in enumerate(res.json()['data'])
        # fmt: on
    ]


@MaxRetry(2)
def get_acg_video_playurl(
    avid: str = "",
    bvid: str = "",
    cid: str = "",
    quality: int = 127,
    audio_quality: int = 30280,
    type: str = "dash",
):
    if not (avid or bvid):
        raise ArgumentsError("avid", "bvid")
    video_quality_sequence = gen_quality_sequence(quality, type=Media.VIDEO)
    audio_quality_sequence = gen_quality_sequence(audio_quality, type=Media.AUDIO)
    play_api = (
        "https://api.bilibili.com/x/player/playurl?avid={avid}&bvid={bvid}&cid={cid}&qn={quality}&type=&otype=json"
    )
    if type == "flv":
        touch_message = spider.get(play_api.format(avid=avid, bvid=bvid, cid=cid, quality=80), timeout=3).json()
        if touch_message["code"] != 0:
            raise CannotDownloadError(touch_message["code"], touch_message["message"])

        accept_quality = touch_message["data"]["accept_quality"]
        for quality in video_quality_sequence:
            if quality in accept_quality:
                break

        play_url = play_api.format(avid=avid, bvid=bvid, cid=cid, quality=quality)
        res = spider.get(play_url, timeout=3)

        return [
            {
                "id": i + 1,
                "url": segment["url"],
                "mirrors": segment["backup_url"],
                "quality": quality,
                "height": video_quality_map[quality]["height"],
                "width": video_quality_map[quality]["width"],
                "size": segment["size"],
                "type": "flv_segment",
            }
            for i, segment in enumerate(res.json()["data"]["durl"])
        ]
    elif type == "dash":
        result = []
        play_api_dash = play_api + "&fnver=0&fnval=2000&fourk=1"
        touch_message = spider.get(
            play_api_dash.format(avid=avid, bvid=bvid, cid=cid, quality=video_quality_sequence[0]), timeout=3
        ).json()

        if touch_message["code"] != 0:
            raise CannotDownloadError(touch_message["code"], touch_message["message"])
        if touch_message["data"].get("dash") is None:
            raise UnsupportTypeError("dash")

        video_accept_quality = set([video["id"] for video in touch_message["data"]["dash"]["video"]])
        for video_quality in video_quality_sequence:
            if video_quality in video_accept_quality:
                break
        else:
            video_quality = 127

        audio_accept_quality = set([audio["id"] for audio in touch_message["data"]["dash"]["audio"]])
        for audio_quality in audio_quality_sequence:
            if audio_quality in audio_accept_quality:
                break
        else:
            audio_quality = 30280

        res = spider.get(play_api_dash.format(avid=avid, bvid=bvid, cid=cid, quality=quality), timeout=3)

        if res.json()["data"]["dash"]["video"]:
            videos = res.json()["data"]["dash"]["video"]
            for video in videos:
                if video["id"] == video_quality:
                    result.append(
                        {
                            "id": 1,
                            "url": video["base_url"],
                            "mirrors": video["backup_url"],
                            "quality": video_quality,
                            "height": video["height"],
                            "width": video["width"],
                            "size": touch_url(video["base_url"], spider)[0],
                            "type": "dash_video",
                        }
                    )
                    break
        if res.json()["data"]["dash"]["audio"]:
            audios = res.json()["data"]["dash"]["audio"]
            for audio in audios:
                if audio["id"] == audio_quality:
                    result.append(
                        {
                            "id": 2,
                            "url": audio["base_url"],
                            "mirrors": audio["backup_url"],
                            "quality": audio_quality,
                            "height": None,
                            "width": None,
                            "size": touch_url(audio["base_url"], spider)[0],
                            "type": "dash_audio",
                        }
                    )
                    break
        return result
    elif type == "mp4":
        play_api_mp4 = play_api + "&platform=html5&high_quality=1"
        play_info = spider.get(play_api_mp4.format(avid=avid, bvid=bvid, cid=cid, quality=127), timeout=3).json()
        if play_info["code"] != 0:
            raise CannotDownloadError(play_info["code"], play_info["message"])
        return [
            {
                "id": 1,
                "url": play_info["data"]["durl"][0]["url"],
                "mirrors": [],
                "quality": play_info["data"]["quality"],
                "height": video_quality_map[play_info["data"]["quality"]]["height"],
                "width": video_quality_map[play_info["data"]["quality"]]["width"],
                "size": play_info["data"]["durl"][0]["size"],
                "type": "mp4_container",
            }
        ]
    else:
        raise UnknownTypeError(type)


@MaxRetry(2)
def get_acg_video_subtitle(avid: str = "", bvid: str = "", cid: str = ""):
    if not (avid or bvid):
        raise ArgumentsError("avid", "bvid")
    subtitle_api = "https://api.bilibili.com/x/player.so?id=cid:{cid}&aid={avid}&bvid={bvid}"
    subtitle_url = subtitle_api.format(avid=avid, cid=cid, bvid=bvid)
    res = spider.get(subtitle_url, timeout=3)
    subtitles_info = json.loads(re.search(r"<subtitle>(.+)</subtitle>", res.text).group(1))
    return [
        # fmt: off
        {
            "lang": sub_info["lan_doc"],
            "lines": spider.get("https:" + sub_info["subtitle_url"], timeout=(3, 10)).json()["body"]
        }
        for sub_info in subtitles_info["subtitles"]
        # fmt: on
    ]



import os
from typing import Optional

from ..utils.base import Text


class Playlist(Text):
    """ 播放列表类 """

    def __init__(self, path: str, path_type: str):
        super().__init__(path)
        self.path_type = path_type

    def switch_path(self, path: str) -> str:
        path = os.path.normpath(path)
        if self.path_type == "AP":
            path = os.path.abspath(path)
        elif self.path_type == "RP":
            path = os.path.relpath(path, start=os.path.dirname(self.path))
        return path

    def write_path(self, path: str) -> None:
        path = self.switch_path(path)
        self.write_string(path)


class M3u(Playlist):
    """ m3u 播放列表 """

    def __init__(self, path: str, path_type: str = "RP"):
        super().__init__(path, path_type)


class Dpl(Playlist):
    """ potplayer 播放列表 """

    def __init__(self, path: str, path_type: str = "RP"):
        super().__init__(path, path_type)
        self.write_string("DAUMPLAYLIST\n")
        self._count = 0

    def write_path(self, path: str, name: Optional[str] = None):
        self._count += 1
        path = self.switch_path(path)
        self.write_string("{}*file*{}".format(self._count, path))
        if name is not None:
            self.write_string("{}*title*{}\n".format(self._count, name))



import os

from .handlers.status import DownloaderStatus
from .quality import audio_quality_map, video_quality_map
from .tools import global_status
from .utils.base import repair_filename
from .utils.console.logger import Logger
from .utils.console.colorful import colored_string


class BililiContainer:
    """bilibili 媒体容器类
    即 B 站上的单个视频，其中可能包含多个媒体单元
    * 包含多个 flv 片段
    * 包含 m4s 的视频与音频流
    * 包含完整的一个 mp4
    """

    def __init__(self, id, name, meta, type="dash", video_dir=""):

        self.id = id
        self.name = name
        self.meta = meta
        self.type = type
        self.path = os.path.join(video_dir, "{}.mp4".format(repair_filename(self.name)))

        self.medias = []
        self.quality = None
        self.height = None
        self.width = None
        self._ = DownloaderStatus(parent=global_status)

    def append_media(self, *args, **kwargs):
        self.medias.append(BililiMedia(*args, **kwargs, container=self))

    def __str__(self):
        quality_description: str = ""
        if self.type == "dash":
            quality_description = " & ".join(
                # fmt: off
                [
                    {
                        "dash_video": video_quality_map,
                        "dash_audio": audio_quality_map
                    }[media.type][media.quality]["description"]
                    for media in self.medias
                ]
                # fmt: on
            )
        else:
            assert self.quality is not None, "quality 仍然为 None"
            quality_description = video_quality_map[self.quality]["description"]
        return "{} 「{}」".format(
            colored_string(self.name, fore="magenta"), colored_string(quality_description, fore="cyan")
        )

    def check_needs_download(self, overwrite: bool = False):
        """ 检查是否需要下载 """
        if overwrite:
            if os.path.exists(self.path):
                os.remove(self.path)
            return True
        if os.path.exists(self.path):
            return False
        return True


class BililiMedia:
    """bilibili 媒体单元类
    从 B 站直接获取的可下载的媒体单元，可能是 flv、mp4、m4s
    """

    def __init__(
        self,
        id,
        url,
        quality,
        size,
        height,
        width,
        container,
        mirrors=[],
        type="dash_video",
        block_size=0,
    ):

        self.id = id
        self.quality = quality
        self.height = height
        self.width = width
        self.url = url
        self.mirrors = mirrors
        self.container = container
        self.block_size = block_size
        self.path = os.path.splitext(self.container.path)[0]
        self.type = type
        if self.container.type == "flv":
            self.path += "_{:02d}.flv".format(id)
        elif self.container.type == "dash":
            self.path += "_{}.m4s".format(type)
        elif self.container.type == "mp4":
            self.path += "_dl.mp4"
        else:
            Logger.warning("未知的容器类型：{}".format(self.container.type))
        self.name = os.path.split(self.path)[-1]
        self._ = DownloaderStatus(parent=self.container._)
        self._.total_size = size

        if self.container.quality is None:
            self.container.quality = quality
        if self.container.width is None:
            self.container.width = width
        if self.container.height is None:
            self.container.height = height
        if self._.total_size == 0:
            Logger.warning("{} 获取 size 为 0".format(self.name))
            self._.total_size = 0
        if self._.total_size is None:
            Logger.warning("{} 无法获取 size".format(self.name))
            self._.total_size = 0

        self.blocks = self.chunking()

    def chunking(self):
        block_size = self.block_size
        blocks = []
        total_size = self._.total_size
        if block_size:
            block_range_list = [(i, i + block_size - 1) for i in range(0, total_size, block_size)]
            if total_size % block_size != 0:
                block_range_list[-1] = (
                    total_size // block_size * block_size,
                    total_size - 1,
                )
        else:
            block_range_list = [(0, total_size - 1)]
        for i, block_range in enumerate(block_range_list):
            blocks.append(
                BililiBlock(
                    id=i,
                    url=self.url,
                    mirrors=self.mirrors,
                    media=self,
                    block_size=block_size,
                    range=block_range,
                )
            )
        assert self._.total_size == total_size, "重新设置的 total size 与原来值不匹配"
        return blocks

    def check_needs_download(self, overwrite=False):
        """ 检查是否需要下载 """
        if overwrite:
            if os.path.exists(self.path):
                os.remove(self.path)
            return True
        if os.path.exists(self.path):
            return False
        return True


class BililiBlock:
    """bilibili 媒体块类"""

    def __init__(self, id, url, mirrors, media, block_size, range):

        self.id = id
        self.url = url
        self.mirrors = mirrors
        self.block_size = block_size
        self.media = media
        self.range = range
        # 假设最大 10 GB 时所需的位数
        ndigits = 1 if block_size == 0 else len(str(10 * 1024 * 1024 * 1024 // self.block_size))
        self.path = "_{:0{}}".format(self.id, ndigits).join(os.path.splitext(self.media.path))
        self.name = os.path.split(self.path)[-1]
        self._ = DownloaderStatus(parent=self.media._)
        self._.total_size = self.range[1] - self.range[0] + 1

    def check_needs_download(self, overwrite=False):
        """ 检查是否需要下载 """
        if overwrite:
            if os.path.exists(self.path):
                os.remove(self.path)
            if os.path.exists(self.path + ".dl"):
                os.remove(self.path + ".dl")
            return True
        if os.path.exists(self.path):
            return False
        return True



import pytest

from bilili.api.bangumi import (
    get_season_id,
    get_bangumi_title,
    get_bangumi_list,
    get_bangumi_playurl,
    get_bangumi_subtitle,
)
from bilili.api.exceptions import CannotDownloadError


@pytest.mark.api
def test_get_season_id():
    media_id = "28223066"
    assert get_season_id(media_id=media_id) == "28770"


@pytest.mark.api
def test_get_title():
    media_id = "28223066"
    assert get_bangumi_title(media_id=media_id) == "我的三体之章北海传"


@pytest.mark.api
def test_get_list():
    season_id = "28770"
    video_list = get_bangumi_list(season_id=season_id, with_section=False)
    assert video_list[0]["cid"] == "144541892"
    assert video_list[0]["avid"] == "84271171"
    assert video_list[0]["bvid"] == "BV1q7411v7Vd"
    assert video_list[0]["episode_id"] == "300998"


@pytest.mark.api
@pytest.mark.ci_skip
@pytest.mark.parametrize("type", ["flv", "dash"])
def test_get_playurl(type: str):
    avid = "84271171"
    bvid = "BV1q7411v7Vd"
    cid = "144541892"
    episode_id = "300998"
    play_list = get_bangumi_playurl(
        avid=avid,
        bvid=bvid,
        cid=cid,
        episode_id=episode_id,
        quality=120,
        audio_quality=30280,
        type=type,
    )


@pytest.mark.api
def test_get_subtitle():
    # TODO: 暂未找到需要字幕的番剧（非港澳台）
    pass



from ..tools import spider
from .utils import MaxRetry


@MaxRetry(2)
def get_danmaku(cid: str) -> str:
    danmaku_api = "http://comment.bilibili.com/{cid}.xml"
    res = spider.get(danmaku_api.format(cid=cid), timeout=(3, 18))
    res.encoding = "utf-8"
    return res.text



from typing import List

noop = lambda *args, **kwargs: None


class Handler:
    """事件处理器"""

    def __init__(self, events: List[str] = []):
        """初始化事件处理器

        Args:
            events (List[str], optional): 事件名称. Defaults to [].
        """
        self.events = events
        for event in self.events:
            setattr(self, event, noop)

    def on(self, event: str):
        """事件添加装饰器

        Args:
            event (str): 事件名称

        Returns:
            function: 在发生 event 后的触发事件，会自动注册在 handler 上
        """
        assert event in self.events

        def on_event(func):

            setattr(self, event, func)

        return on_event
